/*
 * KyuurekiAlmanac.java
 *
 * Created on 2008/01/26, 7:30
 *
 * ----------
 * 
 * Copyright (C) 2007-2011 Yoshitaka Ohsawa All rights reserved.
 * This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 * 
 * ----------
 */

package to.tetramorph.kyuureki;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;
import static java.util.Calendar.*;
import java.util.Date;
import to.tetramorph.time.IntegerDate;
import to.tetramorph.time.JDay;

/**
 * グレゴリオ暦を定気法による日本の旧暦(天保暦)に変換する暦。<br>
 * 変換可能な範囲は1820-12-02から2100-11-30まで。<br>
 * 旧暦は新月の日を暦月の1日目とするため、タイムゾーンの影響を受け、時差が変化すれ
 * ば暦も変化してしまう。たとえば日本で深夜0時から1時に新月を迎えた日、台湾や中国
 * 時間ではまだ前日で、台湾時間で暦を作れば暦月の開始が1日過去の方向にずれて
 * 一ヶ月の日数も1日増えることになる。<br>
 * 　データを増やせばもっと期間を拡張することもできるが、実用範囲はカバーしている
 * と思われる。現在のデータは約30KBである。<br>
 * 　日付をインクリメントしながら参照した場合と、ランダムにした場合ではかなり速度
 * に差がある。これはすべての旧暦データをもっているわけではなく、新月とその暦月の
 * データをもっていて、要求に応じてその月のカレンダーを生成しキャッシュするため
 * で、キャッシュが効率的に機能すれば応答は速くなる。ランダムだとキャッシュは
 * 有効に働かず、そのつどカレンダーを作り直すので効率が低下する。
 * サポートしている期間のうち約55500日分の日付で実験したところ、インクリメントの
 * ときは0.26秒、ランダムのときは9.9秒、デクリメントのときは0.42秒だった。<br>
 * <br>
 * サマータイムによる時差補正はTimeZoneの実装に依存している。
 * 日本で1948-1951年まで実施されたサマータイムは、実装に含まれていない。
 * (と思う、多分)
 * @author 大澤義鷹
 */
public class KyuurekiAlmanac {
    
    private static List<KyuurekiNode> jdayList;  // 新月データ保管リスト
    
    private TimeZone timeZone;                    // 基準とするタイムゾーン

    //新月データから旧暦カレンダーを展開するハッシュ。
    //日時順にアクセスしたい場合もあるのでLinkedHashMapを使用。
    private Map<Integer,KyuurekiDate> map = 
        new LinkedHashMap<Integer,KyuurekiDate>();

    private int min;             //暦の開始日 18201202
    private int max;             //暦の終了日 20321202

    //以下のフィールドはload()実行時にセットされ以後変更されない。

    private static double minJDay;     //暦の開始ユリウス日
    private static double maxJDay;     //暦の終了ユリウス日
    private static double widthJDay;   //開始と終了までのユリウス日の幅
    private static int    rows;        //新月データ件数 jdayList.size()と等しい
    
    /**  
     * デフォルトのタイムゾーンでKyuureki オブジェクトを作成する 
     */
    public KyuurekiAlmanac() {
        if ( jdayList == null ) load();
        setTimeZone(TimeZone.getTimeZone("JST"));
    }
    /**
     * タイムゾーンをセットする。
     */
    protected void setTimeZone( TimeZone timeZone ) {
        this.timeZone = timeZone;
        min = IntegerDate.revjday( minJDay, timeZone );
        max = IntegerDate.revjday( maxJDay, timeZone );
        map.clear();
    }
    /**
     * このオブジェクトが使用しているタイムゾーンを返す。
     */
    public TimeZone getTimeZone() {
        return timeZone;
    }
    /**
     * 新月のユリウス日とその日の暦月のデータをリソースからjdayListに読みこむ。
     * データファイルは約20KB。
     */
    private static void load() {
        jdayList = new ArrayList<KyuurekiNode>();
        DataInputStream dis = null;
        try {
            InputStream is = KyuurekiAlmanac.class.
                getResource("/resources/kyuureki.dat").openStream();
            dis = new DataInputStream(is);
            for ( int i=0; dis.available() != 0; i++ ) {
                double jday = dis.readDouble();
                boolean leap = false;
                int month = dis.readShort();
                if ( month > 12 ) {
                    month = month - 100;
                    leap = true;
                }
                jdayList.add( new KyuurekiNode( leap, month, 1, jday) );
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try { dis.close(); } catch ( Exception e ) { }
        }
        // 旧暦年をナンバリングする
        KyuurekiNode date = jdayList.get(0);
        // 年の位は時差に左右されない
        // (データが正しく冬至中気の新月から始まっていれば)
        Calendar cal = JDay.getCalendar( date.jday, 
                                         TimeZone.getTimeZone("UTC") );
        int year = cal.get(YEAR);
        int month = cal.get(MONTH) + 1;
        for ( int i=0; i<jdayList.size(); i++) {
            KyuurekiNode node = jdayList.get(i);
            if ( node.month == 1 && month == 12) {
                year++;
            }
            node.year = year;
            month = node.month;
        }
        rows = jdayList.size();
        minJDay = jdayList.get(0).jday;
        maxJDay = jdayList.get(rows-1).jday;
        widthJDay = maxJDay - minJDay;
    }
    /**
     * 変換可能なグレゴリオ暦の最後の日付を返す。"20321202"といった値。
     */
    public int getMaximumRange() {
        return max;
    }
    /**
     * 変換可能なグレゴリオ暦の最初の日付を返す。"19201210"といった値。
     */
    public int getMinimumRange() {
        return min;
    }
//    /**
//     * 年月日を"19201210"といった整数表現にして返す。
//     */
//    private static int get(int year,int month,int day) {
//        return year * 10000 + month * 100 + day;
//    }
    /**
     * 指定された新月日のデータから、３ヶ月先までの日毎の旧暦カレンダーを
     * mapに作成する。jdayListの最後のほうが指定され３ヶ月先まで無いときは、
     * あるところまで生成する。過去から未来に向けて順番に生成する。そのとき
     * 無条件に旧暦一ヶ月を30日として生成するが、次の月を生成するとき前回生成
     * された30日目のデータを上書きして、次の月の1日に置き換えていく。
     * この方法だと３ヶ月目の最後の30日目の値は正しい場合もあるし、間違いの場合も
     * ありうる。しかし最後のデータが参照されることはgetKyuurekiDateに限っては
     * ありえない(テスト済み)。
     * ただしgetKyuurekiDateListの場合はありうるので最後のデータは削除する。
     * 同メソッドは値が取得できなくなると、再度getKyuurekiDate()を呼び出す
     * のでそのとき新しくデータが作成される。
     * @param index jdayListの要素番号
     */
    private void makeCal(int index) {
        map.clear();
        int k = index;
        int kmax = index + 3;
        if ( kmax >= rows ) kmax = rows;
        int key = 0;
        for( ; k < kmax; k++ ) {
            KyuurekiNode node = jdayList.get( k );
            Calendar cal = JDay.getCalendar( node.jday, timeZone );
            //System.out.println(String.format("# %tF(%ta)%tT",cal,cal,cal) + "get = " + node.get);

            for(int i=0; i<30; i++) {
                key =   cal.get( YEAR ) * 10000 
                    + ( cal.get( MONTH) + 1 ) * 100 
                    +   cal.get( DAY_OF_MONTH );
                //System.out.println("key = " + key);
                KyuurekiDate kdate = new KyuurekiDate(
                    node.year, node.leap, node.month, node.day + i);
                map.put( key, kdate );
                cal.add( DAY_OF_MONTH, 1 );
            }
        }
        map.remove(key); //一番最後の要素は信用できないので削除
    }
    
    /**
     * グレゴリオ暦を旧暦に変換して返す。時間情報は考慮されない。
     * 時差補正、23時以降の日の繰上、サマータイムによる補正は行わない。
     * @param year 年
     * @param month 月(1-12)
     * @param day 日(1-31)
     * @exception IllegalArgumentException 変換可能な範囲外の日付が指定されたとき。
     */
    public KyuurekiDate getKyuurekiDate(int year,int month,int day) {
        int inputdate = IntegerDate.get( year, month, day );
        if ( inputdate < min ) {
            throw new IllegalArgumentException("暦範囲外 " + inputdate);
        } else if ( inputdate > max ) {
            throw new IllegalArgumentException("暦範囲外 " + inputdate);
        }
        // mapにすでに作成された旧暦カレンダーが存在する場合はそれを返す。
        // 日付をインクリメントしながら連続アクセスする場合にはかなり有効。
        KyuurekiDate cache = map.get( inputdate );
        if ( cache != null ) return cache;
        double jday = JDay.get( year, month, day );
        
        Calendar c = JDay.getCalendar(jday,timeZone);
        //System.out.println(String.format("test --> %tF(%ta)%tT  kdate = %d",c,c,c,inputdate));
        
        double a = (jday - minJDay) / widthJDay;
        int index = (int)(a * rows);
        if ( index > 0 ) index--;
        else if ( index == rows) index = rows-1;
        makeCal(index);
        //dump();
        KyuurekiDate kdate = map.get(inputdate);
        if ( kdate == null ) {
            throw new IllegalArgumentException("暦範囲外 " + inputdate);
        }
        return kdate;
    }
    /**
     * カレンダーで指定されたグレゴリオ暦の年月日を旧暦に変換して返す。
     * 時間情報は考慮されない。
     * getKyuurekiDate(int year,int month,int day) と振る舞いは同じ。
     */
    public KyuurekiDate getKyuurekiDate( Calendar cal ) {
        int y = cal.get( YEAR );
        int m = cal.get( MONTH ) + 1;
        int d = cal.get( DAY_OF_MONTH);            
        return getKyuurekiDate(y,m,d);
    }
    
    private static KyuurekiSupplement kyuurekiSapplement = 
        new KyuurekiSupplement();
    /**
     * カレンダーで指定されたグレゴリオ暦を、サマータイムの実施の有無、
     * 観測地の時差を考慮して時刻補正し旧暦に変換して返す。<br>
     * 紫微斗数などの占いで命盤の作成に必要な旧暦日はこのメソッドを使って求める。
     * <br>
     * 計算の過程でどのような補正が行われたかの情報を取得したい場合は、
     * getKyuurekiDate( Calendar cal,double lon,KyuurekiSupplement sap)を使用する。
     */
    public KyuurekiDate getKyuurekiDate( Calendar cal, double lon ) {
        return getKyuurekiDate( cal, lon, kyuurekiSapplement );
    }
    /**
     * カレンダーで指定されたグレゴリオ暦を、サマータイムの実施の有無、
     * 観測地の時差を考慮して時刻補正し旧暦に変換して返す。<br>
     * 紫微斗数などの占いで命盤の作成に必要な旧暦日はこのメソッドを使って求める。<br>
     * 
     * 観測地の時差の考慮とは、指定された経度を UTCからの時差に変換しこの
     * オブジェクトが使用しているTimeZone#getRawOffset値から引いて差を求める。
     * (この値はマイナスになる事もある)。その差分をカレンダーの日時に加算して
     * 時間を補正する。<br>
     * また時差補正後の時刻が23時以降なら翌日とみなし旧暦月日を返す。<br>
     * sapは、計算した際にどのような日時補正を行ったかの情報を書きこむためのもの。
     * @param cal 日付とタイムゾーンをセットしたカレンダー
     * @param lon 観測地の経度(±0〜180)
     * @param sap 計算の補足情報を返すオブジェクト
     */    
    public KyuurekiDate getKyuurekiDate( Calendar cal, 
                                           double lon,
                                KyuurekiSupplement sap ) {
        sap.isDaylightAdjusted = false;
        sap.isAdjustedNextDay = false;
        sap.adjustCalendar = null;
        sap.adjustTimeByLongitude = 0;
        sap.daylightOffset = 0;
        int offset_ms = (int)(lon / 15 * 3600 * 1000) - timeZone.getRawOffset();
        sap.adjustTimeByLongitude = offset_ms;
        TimeZone tz = cal.getTimeZone();
        int daylightOffset = 0;
        if ( tz.useDaylightTime() ) {      //サマータイム実施がありうる地域なら
            Date nowDate = new Date( cal.getTimeInMillis() );
            if ( tz.inDaylightTime( nowDate ) ) {            //夏時間実施中なら
                sap.isDaylightAdjusted = true;
                daylightOffset = tz.getDSTSavings();
                sap.daylightOffset = daylightOffset;
            }
        }
        Calendar c = (Calendar)cal.clone();
        int adjustOffset = daylightOffset + offset_ms;
        c.add( MILLISECOND, adjustOffset );
        if ( c.get( HOUR_OF_DAY ) >= 23 ) {
            c.add( DAY_OF_MONTH,1 );            
            sap.isAdjustedNextDay = true;
        }
        sap.adjustCalendar = c;
        int y = c.get( YEAR );
        int m = c.get( MONTH ) + 1;
        int d = c.get( DAY_OF_MONTH);
        return getKyuurekiDate(y,m,d);
    }
    /**
     * グレゴリオ暦で指定された日から指定日数分の旧暦データを返す。
     * year,month,dayにはグレゴリオ暦の年月日を指定。月は1-12。
     * @param length 取り出す日数。
     * @return 旧暦日のリスト
     * @exception IllegalArgumentException 暦範囲外の日時やlengthに0以下の値が
     * 入力されたとき。
     */
    public List<KyuurekiDate> getKyuurekiDateList( int year,
                                                    int month,
                                                    int day,
                                                    int length ) {
        if ( length <= 0 ) {
            throw new IllegalArgumentException("lengthは0以上必要");
        }
        Calendar c = Calendar.getInstance();
        c.setLenient(false); //厳密な日付検査を行う
        c.set( year, month - 1, day);
        c.add( DAY_OF_MONTH, length);
        int y = c.get(YEAR);
        int m = c.get(MONTH) + 1;
        int d = c.get(DAY_OF_MONTH);
        int ymd = IntegerDate.get( y, m, d );
        if ( ymd > max ) {
            throw new IllegalArgumentException("暦範囲外" + ymd );
        }
        c.set( year, month - 1, day);
        List<KyuurekiDate> list = new ArrayList<KyuurekiDate>();
        while ( length > 0 ) {
            y = c.get(YEAR);
            m = c.get(MONTH) + 1;
            d = c.get(DAY_OF_MONTH);
            getKyuurekiDate( y, m, d );
            int startdate = IntegerDate.get( y, m, d );
            Iterator ite = map.keySet().iterator();
            int date = 0;
            while ( ite.hasNext() ) {
                date = (Integer)ite.next();
                if ( date != startdate ) continue;
                list.add( map.get( date ) );
                c.add( DAY_OF_MONTH, 1 );
                length--;
                break;
            }
            while ( ite.hasNext() && length > 0 ) {
                date = (Integer)ite.next();
                list.add( map.get(date) );
                c.add( DAY_OF_MONTH, 1 );
                length--;
            }
        }
        return list;
    }
//    //makeCalが生成したデータをダンプする
    void dump() {
        for( Iterator ite = map.keySet().iterator(); ite.hasNext(); ) {
            Integer i = (Integer)ite.next();
            System.out.println( i + "   " + map.get( i ) );
        }
    }
//    //jdayListの内容を先頭からダンプ
//    void test() {
//        for(int i=0; i<10; i++) {
//            KyuurekiNode n = jdayList.get(i);
//            System.out.println(n);
//        }
//    }
    public static void main(String [] args) {
        int year = 1964;
        int month = 9;
        int day = 30;
        int size=33;
        KyuurekiAlmanac k = new KyuurekiAlmanac();
        System.out.println( k.getKyuurekiDate(1964,9,30) );
//        Calendar cal = Calendar.getInstance();
//        cal.set(year,month-1,day);
//        List<KyuurekiDate> klist = k.getKyuurekiDateList(year,month,day,size);
//        int i = 1;
//        for(KyuurekiDate kd : klist) {
//            String gdate = String.format("%02d : %tF(%ta)",i++,cal,cal,cal);
//            System.out.println( gdate + "  →  " + kd );
//            cal.add(DAY_OF_MONTH,1);
//        }
    }
}
