/*
 * SetuiriAlmanac.java
 *
 * Created on 2008/01/17, 23:43
 *
 * ----------
 * 
 * Copyright (C) 2007-2011 Yoshitaka Ohsawa All rights reserved.
 * This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 * 
 * ----------
 */

package to.tetramorph.kyuureki;

import java.io.DataInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.Calendar;
import to.tetramorph.time.JDay;
//import to.tetramorph.starbase.util.JDay;

/**
 * グレゴリオ暦の日時が、その日時の月において節入りしているかどうかの判定を行う暦。
 * <br> 
 * 節入は太陽が各サインの15度にイングレスしたときのことで、通常各月の3〜10日ごろ
 * のあいだに発生する。○年6月1日と入力があったとき、それはたいがい節入り前なの
 * で、判定メソッドはfalseを返す。<br>
 * 正確な節入り時刻はsetuiri.datに年月単位でバイナリ形式で辞書化されており、
 * 約70KBで、リソースからそれを参照する。<br>
 * 判定の有効範囲は1583年1月から2101年2月まで。
 * @author 大澤義鷹
 */
public final class SetuiriAlmanac {
    private static final int MAX = 6300;
    private static int [] dates = new int[MAX];
    private static double [] jdays = new double[MAX]; 
    private static SetuiriAlmanac INSTANCE = null;
    /**
     *  
     * SetuiriAlmanac オブジェクトを作成する。節入暦ファイルをリソースからロードする。
     */
    private SetuiriAlmanac() {
        loadEphemeris();
    }
    
    /**
     * 節入暦ファイルをリソースから取得して配列(dates[],jdays[])に格納する。
     */
    private static void loadEphemeris() {
        DataInputStream dis = null;
        try {
            URL url = SetuiriAlmanac.class.getResource("/resources/setuiri.dat");
            dis = new DataInputStream(url.openStream());
            for ( int i=0; dis.available() != 0; i++ ) {
                dates[i] = dis.readInt();
                jdays[i] = dis.readDouble();
            }
        } catch ( IOException e ) {
            e.printStackTrace();
        } finally {
            try { dis.close(); } catch ( Exception e ) { }
        }
    }
        
    /**
     * 指定したグレゴリオ暦の年月から、その年月の節入り時刻をユリウス日で返す。
     */
    private double setuiri(int year,int month) {
        int key = year * 100 + month;
        int offset = Arrays.binarySearch( dates, key );
        if ( offset < 0 || offset == MAX || key < 158301 || key > 210102) {
            throw new IllegalArgumentException("節入暦範囲外");
        }
        return jdays[ offset ];
    }
    
    /**
     * 指定したカレンダーの年月日時分秒が、その年月の節入りを過ぎたものである
     * 場合はtrueを返す。サポート範囲は1583年1月から2101年2月まで。
     * @exception IllegalArgumentException サポート範囲をこえた日時が指定された時。
     */
    public boolean isSetuiri( Calendar cal ) {
        int year = cal.get( Calendar.YEAR );
        int month = cal.get( Calendar.MONTH ) + 1;
        int day = cal.get( Calendar.DAY_OF_MONTH );
        double inputDay = JDay.get(cal);
        double setuiriDay = setuiri(year,month);
        return inputDay >= setuiriDay;
    }
    /**
     * 指定したグレゴリオ暦の年月から、その年月の節入り時刻を表現するカレンダー
     * を返す。
     */
    public Calendar getSetuiriTime( Calendar cal ) {
        int year = cal.get( Calendar.YEAR );
        int month = cal.get( Calendar.MONTH ) + 1;
        double jd = setuiri( year, month );
        return JDay.getCalendar(jd,cal.getTimeZone());
    }
    /**
     * インスタンスを返す。シングルトンクラスなので、このメソッドでインスタンス
     * を取得する。
     */
    public static SetuiriAlmanac getInstance() {
        if ( INSTANCE == null ) 
            INSTANCE = new SetuiriAlmanac();
        return INSTANCE;
    }

//    public static void main(String [] args) {
//        SetuiriAlmanac setuiri = SetuiriAlmanac.getInstance();
//        for(int i=0; i<10; i++) {
//            System.out.println(setuiri.dates[i]);
//        }
//        System.out.println( setuiri.setuiri(1964,9) );
//        
//    }
}
