/*
 * DateFormat.java
 *
 * Created on 2008/02/19, 13:33
 *
 * ----------
 * 
 * Copyright (C) 2007-2011 Yoshitaka Ohsawa All rights reserved.
 * This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 * 
 * ----------
 */

package to.tetramorph.time;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;

/**
 * カレンダーの値をフォーマットして返すメソッド群。
 * @author 大澤義鷹
 */
public class DateFormat {
    
    private DateFormat() {
    }
    /**
     * グレゴリアンカレンダーのERA情報つきの文字列表現を返す。
     * 例 "AD 2008-02-19(火) 13:55:13"<br>
     * 曜日情報はカレンダーのロケールに基づく。<br>
     */
    public static String getDateString( GregorianCalendar gcal ) {
        String era =
            ( gcal.get( GregorianCalendar.ERA ) == GregorianCalendar.AD ) ?
            "AD"  :  "BC";
        return String.format("%s %tF(%ta) %tT",era, gcal, gcal, gcal );
    }
    /**
     * ユリウス日の日付表現を返す。
     */
    public static String getDateString(double jday,TimeZone tz) {
        return getDateString( JDay.getCalendar( jday, tz ) );
    }
    /**
     * デフォルトタイムゾーンにおけるユリウス日に対応する日付表現を返す。
     */
    public static String getDateString(double jday) {
        return getDateString( JDay.getCalendar( jday, TimeZone.getDefault() ));
    }
    
    /**
     * Calendarの日付文字列を"yyyy-mm-dd,曜日,hh:mm"という書式で返す。<br>
     * 例 "2008-02-19,火,13:39"<br>
     * このメソッドは紀元前はサポートしない。時刻は24時間制。<br>
     * ミリセコンドの桁が500ms以上は切り上げ、次に秒の桁が30秒以上なら切り上げ
     * した値を返す。
     */
    public static String getDateHourMinute( Calendar cal ) {
        //四捨五入する際カレンダーの値を変更してしまうのでクローンを作成
        Calendar gcal = (Calendar)cal.clone();
        //0.5秒以上は切り上げ、以下なら切り捨て
        if ( gcal.get( Calendar.MILLISECOND ) >= 500 ) {
            gcal.add( Calendar.SECOND, 1 );
        }
        gcal.set( Calendar.MILLISECOND, 0 );
        //30秒以上は切り上げ、以下なら切り捨て。
        if ( gcal.get( Calendar.SECOND ) >= 30 ) {
            gcal.add( Calendar.MINUTE, 1 );
        }
        gcal.set(Calendar.SECOND,0);
        return String.format( "%tF,%ta,%tR", gcal, gcal, gcal);
    }

//    public static void main (String [] args) {
//        GregorianCalendar cal = new GregorianCalendar();
//        System.out.println(getDateString(cal));
//    }
}
