/*
 * JDay.java
 *
 * Created on 2006/07/14, 19:05
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 * ----------
 * 
 * Copyright (C) 2007-2011 Yoshitaka Ohsawa All rights reserved.
 * This is free software with ABSOLUTELY NO WARRANTY.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 * 
 * ----------
 */

package to.tetramorph.time;

import java.util.HashMap;
import java.util.Map;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import static java.util.GregorianCalendar.*;

/**
 * 文字列表現の日付・時間・タイムゾーンから、ユリウス日を返すクラス。
 * ユリウス日を求める場合はつねにこのクラスを利用しスイスエフェメリスのSweDateは
 * 使わない。<br>というのはユリウス日は天体や暦計算にはよく使用するので、
 * スイスエフェメリスに依存していると、ライセンスの問題等で不便がある。<br>
 * プログラムを単純化できるという理由から、ユリウス日を求める公式は使用せず、
 * カレンダーのミリ秒表現から直接変換する方法を使っている。
 * TimePlace.getJDay()もこのクラスを呼び出している。
 */
public class JDay {
    //インスタンスは作成不可。
    private JDay() { }

    /**
     * グレゴリオ暦年月日からユリウス日を求めて返す。これはユリウス日を求める
     * 公式を使ったメソッドで、紀元前の値を求めるときは、他のget()メソッドとは
     * 仕様が異なるので注意。ただし多少、パフォーマンスが良いという利点もある。<br>
     * 紀元前1年はyear=0を指定。紀元前2年はyear=-1を指定。GregorianCalendarは
     * 紀元前の年は1年からでERAで紀元前を判定するため、その値をこのメソッドに
     * 渡すときは負数方式の値に変換してから入れなければならない。
     * 時刻の指定はできず、UTCの正午つまり、xx.5の値が返る。
     * @param year グレゴリオ年
     * @param month 1-12を指定。
     * @param day 1-31を指定。
     */
    public static double get(int year,int month,int day) {
        int m = month;
        int y = year;
        if ( month < 3 ) {
            y--;
            m += 12;
        }
        int b = 0;
        if ( (year * 10000 + month * 100 + day) > 15821004 ) {
            int a = y / 100;
            b = 2 - a + (a / 4);
        }
        int c = ( y > 0 ) ? (int)(365.25 * y) : (int)(365.25 * y -0.75);
        return  1720994.5 + (int)( 30.6001 * ( m + 1 ) ) + b + c + day;
    }

    /**
     * 文字列での日時表現からユリウス日を返す。<br>
     * to.tetramorph.starbase.Functionがこのメソッドを使用している。<br>
     * HSQLDBのストアドプロシージャから引数を文字列で引き渡され、それをユリウス日
     * にして返す。そのため場合によっては空白もしくはnullが引き渡される事があり、
     * そのときはnullを返す仕組みになっている。<br><br>
     *
     * ただこのメソッドをこのパッケージに置くのはいかがなものか?<br>
     * 別パッケージに移動したほうがよくないか?<br>
     *
     * @param era "AD"または"BC"を指定する。
     * @param sqlDate "2008-02-22"といったSQLでの日付文字列を指定。
     * @param sqlTime "12:00:00"といったSQLでの時刻文字列を指定。
     * @param tzid "UTC"や"JST"といったTimeZoneクラスがサポートするIDを指定。
     * @param defaultTime 時刻がnullの場合に採用する地方時("hh:mm:dd")
     */
    public static Double get( String era,
                               String sqlDate,
                               String sqlTime,
                               String tzid,
                               String defaultTime ) {
        if ( defaultTime == null || defaultTime.equals("") ) return null;
        if ( sqlDate == null     || sqlDate.equals("") )     return null;
        if ( sqlTime == null     || sqlTime.equals("") ) sqlTime = defaultTime;

        //Timestampで設定すればﾐﾘｾｶﾝﾄﾞの桁もｾﾞﾛになる。
        Timestamp stamp = Timestamp.valueOf(sqlDate + " " + sqlTime);
        //System.out.println("#stamp= " + stamp );

        if ( tzid == null ) tzid = "";
        TimeZone tz = TimeZone.getTimeZone( tzid );

        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime( stamp );

        //cal.setTimeInMillis(stamp.getTime());
        //System.out.printf("#%tF %tT\n",cal,cal);

        if ( era == null ) era = "";
        cal.set( ERA, era.equals("BC") ? BC : AD );

        // tzのセットは最後にしなければならない。そうしないと日時がずれてしまう。

        cal.setTimeZone( tz );
        //System.out.printf("* %tF %tT\n",cal,cal);
        return get( cal );
    }
//    /**
//     * 望んだ日付のカレンダーを簡単に取得できる。
//    上のget()ではTimestampをセットして値をカレンダーにセットしてるのだが、
//    なぜか同じ方法でやってるのに日付がずれる。(UTCを指定したとき)
//     * @param args [ERA],[日付],[時刻]・[タイムゾーン]
//     * @return
//     */
//    public static GregorianCalendar getCalendar( String ... args ) {
//        GregorianCalendar cal = new GregorianCalendar();
//        if ( args == null || args.length == 0 ) {
//            return cal;
//        }
//        Map<String,String> map = new HashMap<String,String>();
//        map.put( "date", String.format( "%tF", cal ));
//        map.put( "time", "00:00:00" );
//        for ( String arg : args ) {
//            if ( arg.matches("^\\d{4}-\\d{2}-\\d{2}")) {
//                map.put( "date", arg );
//            } else if ( arg.matches("^\\d{2}:\\d{2}:\\d{2}")) {
//                map.put( "time", arg );
//            } else if ( arg.matches("AD|BC")) {
//                map.put( "era", arg );
//            } else {
//                map.put( "tz", arg );
//                System.out.println("TZ = " + arg);
//            }
//        }
//        Timestamp stamp = Timestamp.valueOf(
//                map.get("date") + " " + map.get("time") );
//        TimeZone timeZone = TimeZone.getTimeZone( map.get( "tz" ) );
//        System.out.println("#stamp = " + stamp);
//        System.out.println("#tz = " + timeZone.getDisplayName());
//        cal.setTime( Timestamp.valueOf( map.get( "date" ) + " " +
//                                          map.get( "time" ) ) );
//        System.out.printf("#cal = %tF %tT\n", cal,cal);
//
//        if ( map.get("era") != null ) {
//            if ( map.get("era").toUpperCase().equals( "BC" ) ) {
//                cal.set( ERA, BC );
//            }
//        }
//        if ( map.get( "tz" ) != null ) {
//            cal.setTimeZone( TimeZone.getTimeZone( map.get( "tz" ) ) );
//        }
//        System.out.printf("##%tF %tT\n",cal,cal);
//        return cal;
//    }
    /**
     * カレンダーのユリウス日を返す。
     * これは一般的なユリウス日の計算アルゴリズムを使わず、システムのシリアル
     * タイムの規約に従って求める。つまり1970年1月1日UTCを基点に0として、
     * ミリ秒の単位で未来にインクリメント、過去にデクリメントしていく時刻系を
     * ユリウス日に変換する。非常にシンプルにユリウス日を求める事ができる。
     * スイスエフェメリスのユリウス日取得メソッドとBC5400〜AD5400まで値を比較
     * したが、ほとんどの場合、同じ値を取得できる。
     * （１）注意すべきことは、カレンダーはミリ秒の精度をもっている。
     * Calendar.getInstance()で現在時刻を取得したとき、ミリ秒の桁にも値がはいって
     * いて、年月日時分秒までセットしてこのメソッドでユリウス日に変換したとき、
     * ミリ秒の桁の値もユリウス日に反映される。
     * 普通のユリウス日変換は、たいがい秒の桁までしか引数を取らないから、その値と
     * 比べると、わずかにずれていると感じるかもしれない。しかしそれはミリ秒桁を
     * リセットしていないのが原因。
     * （２）カレンダークラスは日本が1948年から1951年まで実施していた、
     * サマータイムを認識している。次の例ではきっかり1時間遅れている。
     * <pre>
     * AD 1948-05-02(日) 23:32:09
     *    millis = -683720871000
     *    sd = 2432674.1056597224
     *    md = 2432674.0639930554,  sd - md = 0.041666666977107525
     * </pre>
     * ところが、Calendarクラスには、それを判定するメソッドが用意されていないのと、
     * TimeZone.useDaylightTimeは、この古い夏時間を認識しない。暗黙のうちに、
     * サマータイムの期間は1時間ずらして日時が求まる仕組みになっている。
     * 今いちよくわかっていないのだが、これは天体位置計算では要注意である。
     * （３）ミリ秒桁がゼロになっている場合、ほとんどの場合はSweDate.getJulDayと
     * 一致するが、一致しない場合のもう一つの例は次のようなもの。
     * <pre>
     * BC 4669-11-03(土) 06:53:34
     *    millis = -209451751586000
     *    sd = 16377.412199074075
     *    md = 16377.412199074073
     * </pre>
     * このメソッドで夏時間に切り替わる期間を狙って、ユリウス日を求めてみると、
     * 同じ０時でもユリウス日の値が変わる事がわかる。
     * <pre>
     * JST0時におけるユリウス日
     * 1948-05-01 00:00:00,2432672.125000
     * 1948-05-02 00:00:00,2432673.125000 (01:00から夏時間に切り替わる)
     * 1948-05-03 00:00:00,2432674.083333 (夏時間になった次の日、jdはずれる)
     * 1948-05-04 00:00:00,2432675.083333
     * </pre>
     * 末尾の桁でわずかに値が異なる。これは計算誤差によるものだが、この桁だと
     * ピコ秒単位での違いでしかなくまったく問題にならない。
     * <pre>
     * 日本でのサマータイム実施は1949-1951年でCalendar.getInstance()などで取得した
     * デフォルトのタイムゾーンのカレンダーに、この期間の日付をセットしてを渡すと、
     * このメソッドは1時間の時差を考慮した値をこのメソッドは返す。
     * 1948-05-02(日) 01:00   1948-09-12 00:00(日)
     * 1949-04-03(日) 01:00   1949-09-11 00:00(日)
     * 1950-05-07(日) 01:00   1950-09-10 00:00(日)
     * 1951-05-06(日) 01:00   1951-09-09 00:00(日)
     *  </pre>
     * 日本における天体計算では、このような夏時間は邪魔になることが多い。これを回避
     * するためにはTimeZoneをUTCで取得後、setRawOffsetで32400000をセットし、
     * そのオブジェクトをCalendarにセットすればいい。
     */

    public static double get( Calendar cal ) {
        double JD0 = 2440587.5; //1970年1月1日 UTC
        long ms = cal.getTimeInMillis();
        long dayms = 86400000L;
        long day = ms / dayms;
        double hour = ms % dayms;
        double jd = day + JD0 + (hour / dayms);
        return jd;
    }

    /**
     * グレゴリオ暦の日時とタイムゾーンからユリウス日を返す。
     * @param era GregorianCalendar.ADまたはBC。
     * @param year 年
     * @param month 月(0-11)
     * @param day 日
     * @param hour 時
     * @param minute 分
     * @param second 秒
     * @param tz TimeZoneオブジェクト
     */
    public static double get( int era,
                                 int year, int month,  int day,
                                 int hour, int minute, int second,
                                 TimeZone tz ) {
        GregorianCalendar cal = new GregorianCalendar(tz);
        cal.set( ERA, era );
        cal.set( Calendar.MILLISECOND, 0);
        cal.set( year, month, day, hour, minute, second );
        return get( cal );
    }


    /**
     * ユリウス日をグレゴリオ暦に変換する。年月日の月は1〜12。
     *
     * @param jday ユリウス日
     * @param tz   タイムゾーン
     */
    public static GregorianCalendar getCalendar( double jday, TimeZone tz ) {
        double JD0 = 2440587.5; //1970年1月1日 UTC
        long millis=(long)(( jday - JD0 ) * 24L * 3600L * 1000L );
        GregorianCalendar c = new GregorianCalendar(tz);
        c.setTimeInMillis(millis);
        return c;
    }
    /**
     * ユリウス日をカレンダーにセットする。
     * カレンダーのインスタンスに、ユリウス日を与えて時刻を変更したい場合に使用する。
     *
     * @param jday セットするユリウス日
     * @param cal セットされるカレンダー
     */
    public static void getCalendar( double jday, Calendar cal ) {
        double JD0 = 2440587.5; //1970年1月1日 UTC
        long millis=(long)(( jday - JD0 ) * 24L * 3600L * 1000L );
        cal.setTimeInMillis(millis);
    }

    /**
     * 夏時間判定をしない日本のTimeZoneオブジェクトを返す。
     */
    public static TimeZone getJSTZone() {
        TimeZone tz = TimeZone.getTimeZone("UTC");
        tz.setRawOffset(32400000);
        tz.setID("JST/UTC-9");
        return tz;
    }

    /**
     * 夏時間のない日本時間のカレンダーを返す。返却されるカレンダーのタイムゾーンは
     * 内部的に作成されたものがセットされ、ゾーンIDは"JST/UTC-9"である。
     * ロケールはデフォルトの設定にまかせる。
     * @return 夏時間の補正のない、日本時間とロケールをもったカレンダー。
     */
    public static Calendar getJSTCalendar( int year,
                                             int month,
                                             int day,
                                             int hour,
                                             int minute,
                                             int second ) {
        TimeZone tz = getJSTZone();
        //Locale loca = new Locale("ja","JP"); //言語コード,国コード
        Calendar cal = Calendar.getInstance(tz);
        cal.set( year, month, day, hour, minute, second );
        cal.set( Calendar.MILLISECOND, 0 );
        return cal;
    }

    /**
     * 夏時間のない日本時間のカレンダーを返す。返却されるカレンダーのタイムゾーンは
     * 内部的に作成されたものがセットされ、ゾーンIDは"JST/UTC-9"である。
     * ロケールはnew Locale("ja","JP")で作成されたものがセットされている。
     * @return 夏時間の補正のない、日本時間とロケールをもったカレンダー。
     */
    public static Calendar getJSTCalendar( int year,
                                             int month,
                                             int day ) {

        return getJSTCalendar( year, month, day, 0, 0, 0 );
    }

//    public static void main(String [] args) {
//        //GregorianCalendar cal = new GregorianCalendar(1951,9-1,3,3,3,0);
//        //GregorianCalendar cal = new GregorianCalendar(1964,9-1,30,5,35,0);
//        GregorianCalendar cal = new GregorianCalendar();
//        cal.set(1951,9-1,3,0,0,0);
//        cal.set( Calendar.MILLISECOND, 0);
//        System.out.println("jday = " + get(cal));
//    }

}
